
package provide polyterm 1.0

namespace eval polyterm {
    namespace export terminal senddev

    # Each polyterm is described by an "instance" array.
    # The instance array names are 0, 1, ...:
    variable Instances		; # Instances(xxx) is set if xxx is a
    				; # current instance.
    variable NextInstance 0	; # Next number to assign to a new polyterm.

    variable OurXTags		; # Tags we've defined, that we care about.
    				; # These are supposed to be mutually exclusive!

    variable this		; # Used by eval_code to store current instance.
				; # Purpose: user code can find instance.
    				; # Beware: multiple windows can modify this
				; # this during execution, if event loop is
				; # processed during eval_code.
#
# Use:
#	set instance [polyterm::terminal  devicepath  name  parent ?options? ]
#
#	Devicepath is one of
#	a) "", meaning to use a dialog to select a device to open;
#	b) host:port, meaning to open a tcp connection to that host and port;
#	c) traffic://host[:port][/wks], meaning to use music messages via
#	    a traffic daemon running on <host>, at port <port>, under
#	    well-known service <wks>.
#	c) traffic://<uta>, ie the Uniform Traffic Address generalization
#	    of the host:port/wks.  Example: traffic://(deirot)/deirot
#	    means traffic server of deirot, target client deirot.
#	d) an actual device pathname.
# Returned:
#	A new container frame (not a megawidget at all), inside of which
#	are a bunch of widgets implementing the terminal.  The frame
#	is named $parent.$NNN, where NNN is the polyterm instance number,
#	and is also the name of an array, polyterm::$NNN, containing all
#	the information needed to manage this terminal.  The caller should
#	pack the frame, and that's it.  The return value is the instance
#	number, NNN.
#
# Options:
#	-bits n[,m]
#		Specifies the number of data and stop bits; valid only for
#		serial lines.  n=data bits (5,6,7,8); m=stop bits (1,2).
#		See also -speed and -parity.
#
#	-buttcols nnn
#		Number of columns of command buttons.
#
#	-buttonlist {name value ...}
#		When a button is pressed, the text is sent to the device.
#		You can define as many buttons as you want.
#
#	-catchputs bool
#		If true, user-entered "puts" to stdout will be caught and
#		redirected into the terminal display.
#
#	-columns n
#		number of columns caller requests for the text widget.
#
#	-crnl bool
#		when stuffing text into the text widget, replace \r\n with \n,
#		and then replace remaining \r with \n.
#
#	-discard {c nnn}
#		Whenever char c is encountered in a text string being sent to
#		the device, it is discarded and an nnn-ms delay is inserted.
#
#	-insertlist {name value ...}
#		When an insert-button is pressed, the text will be inserted
#		into the text-entry box.  You can define as many buttons
#		as you want.
#
#	-lines n
#		number of lines caller requests for the text widget.
#
#	-menubar bool
#		If true, attach the menubar to the toplevel.
#
#	-namespace name
#		Namespace inside which user-entered Tcl code is evaluated
#		(default global).
#
#	-quit bool
#		If true, create a "quit" button that exits the entire
#		application (default false).
#
#	-parity ppp
#		Specifies the parity to be used; valid only for serial lines.
#		ppp may be one of none, odd, even, mark, or space.
#
#	-speed nnn
#		Specifies the baud rate; valid only for serial lines.
#		See also -parity and -bits.
#
#	-termination   none | cr | crlf | lf
#		none: supply no line terminator after user-entered commands;
#		otherwise: use the specified line terminator (see tcl
#		"fconfigure" documentation for details).
#
#	-ticols nnn
#		Number of columns of text-insertion buttons (not cmd buttons).
#
#	-upcase bool
#		Force all user-entered text to uppercase, except quoted text.

    # Each instance array has these elements:
    #	(Note: elements -xxx have a value from some option -xxx.
    #	Elements +xxx are related to some option, but don't have the
    #	directness of a -xxx element.  Elements w/o - or + aren't
    #	closely related to an option.)

	# -catchputs		true if user-entered "puts" to stdout
	#			should be directed into the terminal display.
	# -columns		the requested columns in the text widget
	# -crnl			replace \r\n with \n, when stuffing text widget,
	#			then replace remaining \r with \n.
	# +databits		One of {5,6,7,8}; specifies the number of
	#			data bits for a serial line.
	#			Default: whatever value was set when the
	#			device was opened.  Only valid for serial lines.
	#			Derived from the -bits option.
	# -devtype		If not "", it indicates the device type,
	#			and is used to select default settings.
	# -lines		the requested lines in the text widget
	# -menubar		bool: attach menu to toplevel.
	# -quit			bool: supply a quit button.
	# -parity		One of n,o,e,m,s (none, odd, even, mark, space).
	#			Default: whatever value was set when the
	#			device was opened.  Only valid for serial lines.
	# -speed		Baud rate.
	#			Default: whatever value was set when the
	#			device was opened.  Only valid for serial lines.
	# +stopbits		One of {1,2}; specifies the number of
	#			stop bits for a serial line.
	#			Default: whatever value was set when the
	#			device was opened.  Only valid for serial lines.
	#			Derived from the -bits option.
	# -termination		binary | lf | cr | crlf
	# -upcase		true if typing forced to be uppercase.
	# +use_termination	bool: add line term. after typing or button.
	#			Derived from the -termination option.
	#
	# parent		The window inside which is the new instance.
	# path			pathname to the device, or
	#			hostname:portnum for a networked device,
	#			or traffic://host[:port]/wks for a traffic conn.
	# uta			Iff using traffic, this is the
	#			"uniform traffic address" of the target device.
	# traffic+port		Iff using traffic, this is the host:port number
	#			of the traffic server.
	# my_music_addr	Iff using traffic, this is our music address.
	#			Otherwise, it is set to -1.
	# -msg			For use with MUSIC messages.  The value is a
	#			message number; the message body is the raw
	#			text to/from the device.
	#			The "data1" field in the header contains:
	#			  -1: message is from device;
	#			  otherwise: music message address of sender.
	#			    (If addr is our own, we ignore the message).
	# -devnum		For use with MUSIC messages.  If there are
	#			multiple devices controlled by a dispatcher,
	#			use this number to select the device.
	#			(It is passed in the buf[0] location of
	#			the header.)
	# opendialog		bool: use dialog to select device.
	# name			descriptive name for this device
	# handle		if dev is open: handle; otherwise: "".
	# filemenu		the File menu.
	# openitem		the index of the Open entry in the File menu.
	# closeitem		the index of the Close entry in the File menu.
	# text			the text widget
	# timeout		after a socket open attempt, the amount of
	#			time (millisec) before we give up on the
	#			connection attempt.
	# memory		a copy of the actual stuff received from
	#			the device.
	# tcl,history		what we've seend from the tcl-entry
	#			to the widget.
	# tcl,histidx		index to where we should get history,
	#			where 0=end, 1=end-1, etc.
	# cmd,history		what we've seend from the command-entry
	#			to the widget.
	# cmd,histidx		index to where we should get history,
	#			where 0=end, 1=end-1, etc.
	# typed,back_subst	bool: typed cmds are backslash-subst'd.
	# typed,var_subst	bool: typed cmds are variable-subst'd.
	# typed,cmd_subst	bool: typed cmds are command-subst'd.
	# button,back_subst	bool: cmd buttons are backslash-subst'd.
	# button,var_subst	bool: cmd buttons are variable-subst'd.
	# button,cmd_subst	bool: cmd buttons are command-subst'd.
	# modestring		Displayable string like "9600 8N1".
	#			Updated whenever we try to adjust the actual
	#			serial settings for the device.
	# statuscmd		append str and execute to set the status line.
	# codeentry		textvariable used for tcl code typing.
	# cmdentry		textvariable used for cmd-line typing.
	# inquote		Boolean: true if inside quote
	# +delayamount,x	For each character x, $instance(DelayAmount,x)
	#			says that char x is replaced with
	#			$instance(DelayAmount,x) ms delay.
	# -insertlist		InsertList is a set of label/text pairs.
	#			When a label is pressed, the text string
	#			is inserted into the entry box.
	# -buttonlist		ButtonList is a set of label/action pairs.
	#			When a button is pressed, the action string
	#			is sent to the device.
	# -buttcols		number of columns of cmd buttons.
	# -ticols		number of columns of text-insertion buttons.
	# -namespace		Namespace in which we evaluate
	#			user-typed code.
	# fromdevformat		How to display data received from
	#			the device: text, binary, octal, hex.
	# $toplev,WaitFor	scratch variable, used by the WaitFor proc
	#			to hold the "after" id for this item.
	# master		Name of the master frame.
	# cmdwin		Name of the frame holding the command entry
	#			area and insertlist text.
	# row,cmdwin		Number of the grid row where cmdwin goes.

proc terminal {devicepath name parent args} {

    set this [create_instance $devicepath $name $parent]

    if {[catch {config_instance $this $args} result]} {
	tk_dialog .d "Configuration Failed." \
		    "Can't configure $devicepath -- $result" error 0 OK
	return ""
    }

    create_windows $this

    return $this
}

#
# A puts emulation for use inside user's tcl code: if output goes to stdout,
# then we capture it and put it into the text widget, instead.
#
# It assumes that true_puts will implement the original puts.
#
proc pt_puts {instance args} {
    variable $instance
    upvar 0 $instance state

    switch [llength $args] {
	1 {
	    set newline 1
	    set channel stdout
	    set string [lindex $args 0]
	}
	2 {
	    foreach {arg0 string} $args {break}
	    if {[string match "$arg0*" -nonewline]} {
		set newline 0
		set channel stdout
	    } else {
		set newline 1
		set channel $arg0
	    }
	}
	3 {
	    foreach {opt channel string} $args {break}
	    if {![string match "$opt*" -nonewline]} {
		return -code error "bad option \"$opt\":\
					should be \"-nonewline\""
	    }
	    set newline 0
	    set string [lindex $args 2]
	}
	default {
	    eval true_puts $args	; # generates error message
	}
    }
    if {[string compare $channel stdout]} {
	if {$newline} {
	    true_puts $channel $string
	} else {
	    true_puts -nonewline $channel $string
	}
    } else {
	if {$newline} {
	    @NS@::stuff_text $instance "$string\n" tclcode
	} else {
	    @NS@::stuff_text $instance "$string" tclcode
	}
    }
}

proc create_instance {path name parent} {
    variable Instances			; # keeps track of polyterms we create.
    variable NextInstance		; # keeps track of polyterms we create.

    variable $NextInstance		; # array with this instance data.
    upvar 0 $NextInstance state		; # for convenience

    set this $NextInstance
    incr NextInstance

    set Instances($this) 1

    if {[string compare $path ""] == 0} {
	set state(opendialog)		1
	set state(path)			""
	set state(uta)			""
	set state(traffic+port)		""
    } else {
	set state(opendialog)		0
	set state(path)			$path
	set state(uta)			""
	set state(traffic+port)		""
    }
    set state(name)			$name
    set state(parent)			$parent

    init_instance $this

    set this
}

proc init_instance {instance {reinit 0}} {
    variable $instance
    upvar 0 $instance state

    set state(+databits)		""
    set state(+stopbits)		""
    set state(+use_termination)		1
    set state(-buttonlist)		{}
    set state(-catchputs)		1
    set state(-columns)			80
    set state(-devtype)			""
    set state(-insertlist)		{}
    set state(-lines)			24
    set state(-menubar)			0
    set state(-namespace)		::
    set state(-parity)			""
    set state(-quit)			0
    set state(-buttcols)		8
    set state(-ticols)			8
    set state(-speed)			""
    set state(-termination)		auto
    set state(-upcase)			0

    # Only do this if it's not a re-init
    if {!$reinit} {
	set state(button,back_subst)	1
	set state(button,cmd_subst)	0
	set state(button,var_subst)	1
	set state(close)		0
	set state(closeitem)		""
	set state(cmdentry)		""
	set state(codeentry)		""
	set state(-crnl)		0
	set state(fromdevformat)	text
	set state(handle)		""
	set state(inquote)		0
	set state(memory)		""
	set state(modestring)		""
	set state(openitem)		""
	set state(text)			""
	set state(cmd,history)		{}
	set state(tcl,history)		{}
	set state(cmd,histidx)		0
	set state(tcl,histidx)		0
	set state(timeout)		3000
	set state(typed,back_subst)	1
	set state(typed,cmd_subst)	0
	set state(typed,var_subst)	0
	set state(-msg)			""
	set state(-devnum)		0
	set state(my_music_addr)	-1
    }

}

proc config_instance {instance optionlist} {
    variable $instance
    upvar 0 $instance state

    # argument list is of form
    #	key value key value ...
    # Each key can be a unique leading substring of actual key.
    set Options {
	-bits
	-buttcols
	-buttonlist
	-catchputs
	-columns
	-crnl
	-devicetype
	-devtype
	-discard
	-insertlist
	-lines
	-menubar
	-namespace
	-parity
	-quit
	-speed
	-termination
	-ticols
	-upcase
	-msg
	-devnum
    }

    foreach {key value} $optionlist {
	set i [lsearch -glob $Options "$key*"]
	if {$i < 0} {
	    return -code error "No such option as $key"
	}
	set opt [lindex $Options $i]
	switch -glob -- $opt {
	    -bits {
		# Data and stop bits: pattern is n[,m]
		if {[regexp {^[5678]$} $value]} {
		    set state(+databits) $value
		} elseif {[regexp {^[5678]$|^[5678],[12]$} $value]} {
		    foreach {state(+databits) \
		    	state(+stopbits)} [split $value ,] {break}
		} else {
		    return -code error "Option -bits needs an argument\
			of the form n\[,m] (n=5..8; m=1,2), but got `$value'"
		}
	    }
	    -buttcols {
		if {![regexp -- {^[1-9][0-9]*$} $value]} {
		    return -code error "Option -buttcols needs a positive\
				number, but got `$value'"
		}
		set state(-buttcols) $value
	    }
	    -buttonlist {
		set state(-buttonlist) $value
	    }
	    -catchputs {
		if {[catch {testbool $value} state(-catchputs)]} {
		    return -code error "Option -catchputs needs a\
		    		boolean argument, but got `$value'"
		}
	    }
	    -columns {
		if {![regexp -- {^[1-9][0-9]*$} $value]} {
		    return -code error "Option -columns needs a positive\
				number, but got `$value'"
		}
		set state(-columns) $value
	    }
	    -crnl {
		if {[catch {testbool $value} state(-crnl)]} {
		    return -code error "Option -crnl needs a\
		    		boolean argument, but got `$value'"
		}
	    }
	    -devnum { set state(-devnum) $value }
	    -devicetype { - }
	    -devtype {
		# Recursively configure ourselves for this device type.
		if {[string compare $value ""] == 0} {
		    # No device type named.
		} else {
		    # First get the canonical name (so that we can
		    # assign it to -devtype); then get the attributes
		    # and pass them to config_instance.
		    # (Note that type_info can generate an error, but we
		    # don't catch it.)
		    set canonical [type_info -canonical $value]
		    # If the type doesn't start with "+", make it into
		    # our devtype:
		    if {![string match +* $canonical]} {
			set state(-devtype) $canonical
		    }
		    set instance_args [type_info -get $canonical]
		    if {[llength $instance_args]} {
			config_instance $instance $instance_args
		    }
		}
	    }
	    -discard {
		set c [lindex $value 0]
		set delay [lindex $value 1]
		if {[string length $c] != 1} {
		    return -code error "Incorrect format for\
			-discard character; should be `-discard c nnn'"
		}
		if {![regexp -- {^[0-9]+$} $delay]} {
		    return -code error "Incorrect form for\
			    delay specifier; should be `-discard c nnn'"
		}
		set state(+delayamount,$c) $delay
	    }
	    -insertlist {
		set state(-insertlist) $value
	    }
	    -lines {
		if {![regexp -- {^[1-9][0-9]*$} $value]} {
		    return -code error "Option -lines needs a positive\
				number, but got `$value'"
		}
		set state(-lines) $value
	    }
	    -menubar {
		if {[catch {testbool $value} state(-menubar)]} {
		    return -code error "Option -menubar needs a\
		    		boolean argument, but got `$value'"
		}
	    }
	    -msg { set state(-msg) $value }
	    -namespace {
		set state(-namespace) $value
	    }
	    -parity {
		if {[string match $value* none]} {
		    set state(-parity) "n"
		} elseif {[string match $value* odd]} {
		    set state(-parity) "o"
		} elseif {[string match $value* even]} {
		    set state(-parity) "e"
		} elseif {[string match $value* mark]} {
		    set state(-parity) "m"
		} elseif {[string match $value* space]} {
		    set state(-parity) "s"
		} else {
		    return -code error "Option -parity accepts odd, even,\
		    	mark, space, or none, but got `$value'"
		}
	    }
	    -quit {
		if {[catch {testbool $value} state(-quit)]} {
		    return -code error "Option -quit needs a\
		    		boolean argument, but got `$value'"
		}
	    }
	    -speed {
		if {![regexp {^[0-9]+$} $value state(-speed)]} {
		    return -code error "Option -speed needs an argument\
					of the form nnn, but got `$value'"
		}
	    }
	    -termination {
		set use 1
		switch -exact -- $value {
		none { set use 0; set state(-termination) binary }
		cr   { set use 1; set state(-termination) cr }
		crlf { set use 1; set state(-termination) crlf }
		lf   { set use 1; set state(-termination) lf }
		default {
		    return -code error "Option -termination accepts types\
		    	none, cr, crlf, or lf, but got `$value'"
		    }
		}
		set state(+use_termination) $use
	    }
	    -ticols {
		if {![regexp -- {^[1-9][0-9]*$} $value]} {
		    return -code error "Option -ticols needs a positive\
				number, but got `$value'"
		}
		set state(-ticols) $value
	    }
	    -upcase {
		if {[catch {testbool $value} state(-upcase)]} {
		    return -code error "Option -upcase needs a\
		    		boolean argument, but got `$value'"
		}
	    }
	    default {
		return -code error "Internal error: option `$key' was\
			matched to option `$opt', but we don't have code\
			to process such an option!"
	    }
	}
    }
}

#
# returns 1 (0) if the value is (is not) true; generates an error if it's
# not a boolean.
#
proc testbool {val} {
    if {"$val"} {return 1} else {return 0}
}

proc create_windows {instance} {
    variable OurXTags
    variable $instance
    upvar 0 $instance state	; # Note that state is now an alias for the
    				; # instance info.

    array set state [subst -nobackslashes -nocommands {
	buttcols	$state(-buttcols)
	UserTypedColor	blue
	CmdButtonColor	red
	CodeColor	darkgreen
	UnknownColor	black
	UnknownFont	{-family Times -slant italic}
    }]
    # For convenience...
    set buttcols	$state(-buttcols)
    set UserTypedColor	$state(UserTypedColor)
    set CmdButtonColor	$state(CmdButtonColor)
    set CodeColor	$state(CodeColor)
    set UnknownColor	$state(UnknownColor)
    set UnknownFont	$state(UnknownFont)
    
    set NS [namespace current]

    # Get the fully-qualified instance array name, for use in callbacks
    # and as widget variables.
    set NSI [namespace which -variable $instance]

    #
    # ----------------------------------------------------------------
    # Create a master frame for this instance, and work inside that frame.
    # Note that if the caller has already created a widget named
    # <parent>.$instance, then this call will fail, hence our careful
    # checking of the frame command.
    #
    if {[catch {frame $state(parent).$instance -relief raised -bd 2} master]} {
	return -code error "Failed to create polyterm master frame: $master"
    }

    # If the master is destroyed, also close the device and destroy the
    # state data.
    set script {
	if {[string compare %W $master] == 0} {
	    ${NS}::cleanup_terminal $instance
	}
    }
    bind $master <Destroy> [subst -nocommands $script]

    set state(master) $master

    set Row -1

    #
    # ----------------------------------------------------------------
    # This row: File menu; Format menu; device status.
    #
    if {$state(-menubar)} {
	# Menubar on toplevel
	set toplevel [winfo toplevel $master]
	set menubar [menu $toplevel.polyterm_menubar$instance]
	$toplevel configure -menu $menubar
	# The funny use of m_file, m_format, etc is so that we can
	# construct variables that work for both menubar and normal frame.
	set m_file $menubar
	$menubar add cascade -label File -menu $m_file.file
	set m_edit $menubar
	$menubar add cascade -label Edit -menu $m_edit.edit
	set m_label $menubar
	$menubar add command -label "xxx" -state disabled -font fixed
	set index [$menubar index last]
	set state(statuscmd) [list $menubar entryconfig $index -label]
	set state(statuscolorcmd) [list $menubar config -disabledforeground]
    } else {
	# Add menubuttons into normal frame.
	incr Row
	set menubar [frame $master.menubar]
	grid configure $menubar -row $Row -column 0 \
					-columnspan $buttcols -sticky ew
	set m_file $menubar.file
	menubutton $m_file -text File -menu $m_file.file
	set m_edit $menubar.edit
	menubutton $m_edit -text Edit -menu $m_edit.edit
	set m_label $menubar.label
	menubutton $m_label -text "xxx" -font fixed
	set state(statuscmd) [list $m_label config -text]
	set state(statuscolorcmd) [list $m_label config -foreground]

	$menubar configure -bg [$menubar.file cget -bg]

	pack $menubar.file $menubar.edit $menubar.label -side left -fill x
    }

    # ++++++++++++++++
    # File menu
    # ++++++++++++++++
    set fm [menu $m_file.file -tearoff 0]
    set state(filemenu) $fm
    if {$state(opendialog)} {
	$fm add command -label Open... -command \
					[list ${NS}::open_dialog $instance]
    } else {
	$fm add command -label Open -command [list ${NS}::open_device $instance]
    }
    set state(openitem) [$fm index last]
    $fm add command -label Comm -command [list ${NS}::setcomm $instance]

    $fm add command -label Close -command [list ${NS}::close_device $instance]
    set state(closeitem) [$fm index last]

    setstatus $instance closed

    if {$state(-quit)} {
	$fm add command -label Quit \
			    -command [list ${NS}::confirm_quit $instance]
    }

    # ++++++++++++++++
    # Edit menu
    # ++++++++++++++++
    set em [menu $m_edit.edit -tearoff 0]

    $em add command -label "Delete Text" -command \
    					[list ${NS}::deletetext ${NSI}]

    $em add check -label Bold -variable ${NSI}(fontweight)
    # N.B.  A trace will be added later to this variable.

    set fs $em.fontslant
    $em add check -label Slanted -variable ${NSI}(fontslant)
    # N.B.  A trace will be added later to this variable.

    set fs $em.fontsize
    $em add cascade -label Size -command \
			[list fontsizename $fs ${NSI}] -menu $fs
    menu $fs -tearoff 0
    $fs add radio -label Points -variable ${NSI}(fontunit) -value pt
    $fs add radio -label Pixels -variable ${NSI}(fontunit) -value px
    $fs add radio -label 8 -variable ${NSI}(fontsize) -columnbreak 1
    $fs add radio -label 9 -variable ${NSI}(fontsize)
    $fs add radio -label 10 -variable ${NSI}(fontsize)
    $fs add radio -label 11 -variable ${NSI}(fontsize)
    $fs add radio -label 12 -variable ${NSI}(fontsize)
    $fs add radio -label 14 -variable ${NSI}(fontsize)
    $fs add radio -label 16 -variable ${NSI}(fontsize)
    $fs add radio -label 18 -variable ${NSI}(fontsize)
    # N.B.  A trace will be added later to the fontunit and fontsize variables.

    set fm $em.format
    $em add cascade -label Format -menu $fm
    menu $fm -tearoff 1
    foreach type {text binary octal hex} {
	$fm add command -label $type -command [list \
			    ${NS}::changeDisplayedType $instance $type]
    }

    #
    # ----------------------------------------------------------------
    # This row: empty spacer, to improve visual appearance.
    #
    incr Row
    set f $master.sep$Row
    frame $f -borderwidth 2
    grid configure $f -row $Row -column 0 -columnspan $buttcols -sticky ewns
    grid rowconfigure $master $Row -minsize 8

    #
    # ----------------------------------------------------------------
    # This row: user-entered Tcl code.
    #
    incr Row
    set code [frame $master.code]
    grid configure $code -row $Row -column 0 -columnspan $buttcols -sticky ew
    set l $code.lab
    label $l -text "Tcl (namespace $state(-namespace)): " -fg $CodeColor
    pack $l -side left

    set e $code.entry
    entry $e -textvariable ${NSI}(codeentry) -fg $CodeColor \
    	-font "courier -12"
    bind $e <Key-Backspace> [bind $e <Key-Delete>]
    bind $e <Key-Return> [list ${NS}::eval_code_and_reset $instance %W]
    bind $e <Up> [list ${NS}::setentryfromhist $instance $e tcl 1]
    bind $e <Down> [list ${NS}::setentryfromhist $instance $e tcl -1]
    pack $e -side right -expand true -fill x

    #
    # ----------------------------------------------------------------
    # This row: user-entered text to send to device.
    # Put it into a frame that is shared with the text-insert buttons.
    #
    incr Row
    set state(cmdwin) "$master.cmdwin"
    set state(row,cmdwin) $Row

    create_cmdwin $instance

    #
    # ----------------------------------------------------------------
    # This row: empty spacer, to improve visual appearance.
    #
    incr Row
    set f $master.sep$Row
    frame $f -borderwidth 2
    grid configure $f -row $Row -column 0 -columnspan $buttcols -sticky ewns
    grid rowconfigure $master $Row -minsize 8

    #
    # ----------------------------------------------------------------
    # This row: the text widget.
    #
    incr Row
    set f [frame $master.textframe]
    grid configure $f -row $Row -column 0 -columnspan $buttcols -sticky ewns
    grid rowconfigure $master $Row -weight 1

    set state(text) $f.text
    set ysb $f.yscroll
    scrollbar $ysb -command [list $state(text) yview] -orient vertical
    text $state(text) -wrap char -height $state(-lines) \
			-bg white \
			-width $state(-columns) -state disabled \
			-yscrollcommand [list $ysb set]

    # The following are mutually-exclusive tags.  If you need to specify
    # additional tags that are active at the same time as these tags,
    # don't add them to the OurXTags list!
    # (OK, so we redefine this list on each call to the proc.  No big deal.)
    set OurXTags {fromdevice usertyped tclcode frombutton fromunknown}
    $state(text) tag configure fromdevice
    $state(text) tag configure usertyped -foreground $UserTypedColor
    $state(text) tag configure tclcode -foreground $CodeColor
    $state(text) tag configure frombutton -foreground $CmdButtonColor
    $state(text) tag configure fromunknown -foreground $UnknownColor \
    				-font $UnknownFont

    grid configure $state(text) -row 0 -column 0 -sticky ewns
    grid configure $ysb -row 0 -column 1 -sticky ns
    grid rowconfigure $f 0 -weight 1

    #
    # ----------------------------------------------------------------
    # This row: the caller-defined buttons.
    #
    incr Row
    set state(buttwin) "$master.buttwin"
    set state(row,buttwin) $Row

    create_buttwin $instance

    #
    # ----------------------------------------------------------------
    # Get the initial values for fontweight, slant, etc.
    #
    array set font [font actual [$state(text) cget -font]]
    set state(fontweight) $font(-weight)
    set state(fontslant) $font(-slant)
    if {$font(-size) > 0} {
	set state(fontsize) $font(-size)
	set state(fontunit) "pt"
    } else {
	set state(fontsize) [expr -$font(-size)]
	set state(fontunit) "px"
    }
    trace variable ${NSI}(fontweight) w ${NS}::setfont
    trace variable ${NSI}(fontslant) w ${NS}::setfont
    trace variable ${NSI}(fontunit) w ${NS}::setfont
    trace variable ${NSI}(fontsize) w ${NS}::setfont
}

proc fontsizename {menu instance} {
    upvar 0 $instance state
    $menu config -label "$state(fontsize) $state(fontunit)"
}

#
# A variable trace to trigger clearing the text buffer.
#
proc deletetext {instance} {
    upvar 0 $instance state

    $state(text) configure -state normal
    $state(text) delete 1.0 end
    $state(text) configure -state disabled

    set state(cmd,history)		{}
    set state(tcl,history)		{}
    set state(cmd,histidx)		0
    set state(tcl,histidx)		0
}

#
# A variable trace (the variable should be on a radiobutton)
# for setting font info.
#
proc setfont {var idx op} {
    upvar 0 $var state

    set overall_fontinfo [$state(text) cget -font]
    set device_fontinfo [$state(text) tag cget fromdevice -font]
    if {[string compare $device_fontinfo ""] == 0} {
	set device_fontinfo $overall_fontinfo
    }

    switch $idx {
	# Only fuss with the weight of the "fromdevice" text
	fontweight {
	    array set font [font actual $device_fontinfo]
	    switch $state(fontweight) {
		Bold { set font(-weight) bold }
		default { set font(-weight) normal }
	    }
	    $state(text) tag config fromdevice -font [array get font]
	}
	fontslant {
	    array set font [font actual $device_fontinfo]
	    switch $state(fontslant) {
		Slanted { set font(-slant) italic }
		default { set font(-slant) roman }
	    }
	    $state(text) tag config fromdevice -font [array get font]
	}
	fontunit -
	fontsize {
	    set size $state(fontsize)
	    if {[string compare $state(fontunit) "px"] == 0} {
		set size -$size
	    }
	    # First do the overall text widget font:
	    array set font [font actual $device_fontinfo]
	    set font(-size) $size
	    $state(text) config -font [array get font]
	    # Then do any fonts associated with specific tags:
	    foreach tag [$state(text) tag names] {
		set fontinfo [$state(text) tag cget $tag -font]
		if {[string compare $fontinfo ""]} {
		    array set font [font actual $fontinfo]
		    set font(-size) $size
		    $state(text) tag config $tag -font [array get font]
		}
	    }
	}
    }
}

#
# Create the command entry area and insertlist text.
#
proc create_cmdwin {instance} {
    variable $instance
    upvar 0 $instance state	; # Note that state is now an alias for the
    				; # instance info.

    set NS [namespace current]
    set NSI [namespace which -variable $instance]

    # For convenience...
    set buttcols	$state(-buttcols)
    set UserTypedColor	$state(UserTypedColor)
    set CmdButtonColor	$state(CmdButtonColor)
    set CodeColor	$state(CodeColor)

    # Create this frame, no matter what, so that it can be a "root-ish"
    # window from which we destroy all cmd-entry and insertlist window info.
    set cmdwin [frame $state(cmdwin)]

    grid configure $cmdwin -row $state(row,cmdwin) \
    				-column 0 -columnspan $buttcols -sticky ew

    set l $cmdwin.cmdlab
    label $l -text "Command: " -fg $UserTypedColor
    grid configure $l -row 0 -column 0 -sticky w

    set e $cmdwin.entry
    entry $e -textvariable ${NSI}(cmdentry) -fg $UserTypedColor \
    	-font "courier -12"
    grid configure $e -row 0 -column 1 -sticky ew
    grid columnconfigure $cmdwin 1 -weight 1

    # Set up checkbuttons for controlling substitution:
    set uc [checkbutton $cmdwin.uc -text "^" -fg $UserTypedColor \
				-selectcolor $UserTypedColor \
    				-variable ${NSI}(-upcase)]
    set bs [checkbutton $cmdwin.bs -text "\\" -fg $UserTypedColor \
				-selectcolor $UserTypedColor \
    				-variable ${NSI}(typed,back_subst)]
    set cs [checkbutton $cmdwin.cs -text "\[ ]" -fg $UserTypedColor \
				-selectcolor $UserTypedColor \
    				-variable ${NSI}(typed,cmd_subst)]
    set vs [checkbutton $cmdwin.vs -text "\$" -fg $UserTypedColor \
				-selectcolor $UserTypedColor \
    				-variable ${NSI}(typed,var_subst)]
    grid configure $uc -row 0 -column 2 -sticky ew
    grid configure $bs -row 0 -column 3 -sticky ew
    grid configure $vs -row 0 -column 4 -sticky ew
    grid configure $cs -row 0 -column 5 -sticky ew

    bind $e <Key-Backspace> [bind $e <Key-Delete>]
    bind $e <Key-Return> [list ${NS}::send_and_reset $instance %W usertyped]
    bind $e <Up> [list ${NS}::setentryfromhist $instance $e cmd 1]
    bind $e <Down> [list ${NS}::setentryfromhist $instance $e cmd -1]

    set state(inquote) 0
    bind $e <Key-quotedbl> [list ${NS}::invert $instance inquote]
    foreach letter [split abcdefghijklmnopqrstuvwxyz ""] {
	# Control characters:
	set script {tkEntryInsert %W "@letter@"; break}
	regsub @letter@ $script [cntl $letter] script
	bind $e <Control-Key-$letter> $script
	# Normal letters:
	set upcase [string toupper $letter]
	set script {
	    if {$@NSI@(-upcase) && !$@NSI@(inquote)} {
		tkEntryInsert %W "@upcase@"; break
	    }
	    # lowercase letters handled by binding falling through 
	    # to standard binding.
	}
	regsub -all @upcase@ $script $upcase script
	regsub -all @NSI@ $script $NSI script
	bind $e <Key-$letter> $script
    }

    #
    # ----------------------------------------------------------------
    # This row: the text-insertion buttons.  Note that they go into a grid
    # that is shared with the "Command" row, above.
    #
    # Put all the insertion buttons into the "1'th" column, ie the space
    # occupied by the command-entry button.   Subdivide this into a series
    # of $inscolumns.
    set inscolumns $state(-ticols)
    if {[llength $state(-insertlist)]} {

	set l $cmdwin.inslab
	label $l -text "Insert: " -fg $UserTypedColor
	grid configure $l -row 1 -column 0 -sticky w

	set f [frame $cmdwin.insert]
	grid configure $f -row 1 -column 1 -columnspan 5 -sticky ew
	# Use an "imaginary" column (ie one that we won't fill) over on the
	# right side to soak up any extra space, and force the other columns
	# to pack to the left.
	grid columnconfigure $f $inscolumns -weight 1

	set col -1
	set row 0
	foreach {label text} $state(-insertlist) {
	    incr col
	    if {$col >= $inscolumns} { incr row ; set col 0 }
	    set butt $f.${row}_${col}
	    button $butt -text $label -fg $UserTypedColor \
		-activeforeground $UserTypedColor \
		-command [list ${NS}::subst_append $instance $e $text] \
		-borderwidth 2 -font {courier -12} -padx 1 -pady 1
	    grid configure $butt -row $row -column $col -sticky nsew
	}
    }
}

#
# Return history info, and bump the history pointer.
#
proc gethist {instance which direction} {
    variable $instance
    upvar 0 $instance state	; # Note that state is now an alias for the
    				; # instance info.
    set n [llength $state($which,history)]
    set idx [expr $n - $state($which,histidx) - $direction] ; # relative to end.
    if {$idx < 0 || $idx >= $n} {
	return ""
    }
    incr state($which,histidx) $direction
    return [lindex $state($which,history) $idx]
}

#
# Set an entry box from a history list.
#
proc setentryfromhist {instance widget which direction} {
    variable $instance
    upvar 0 $instance state	; # Note that state is now an alias for the
    				; # instance info.
    set NS [namespace current]
    set text [${NS}::gethist $instance $which $direction]
    $widget delete 0 end
    $widget insert 0 $text
}

#
# Create the buttonlist buttons.
#
proc create_buttwin {instance} {
    variable $instance
    upvar 0 $instance state	; # Note that state is now an alias for the
    				; # instance info.

    set NS [namespace current]
    set NSI [namespace which -variable $instance]

    # For convenience...
    set buttcols	$state(-buttcols)
    set UserTypedColor	$state(UserTypedColor)
    set CmdButtonColor	$state(CmdButtonColor)
    set CodeColor	$state(CodeColor)

    # Create this frame, no matter what, so that it can be a "root-ish"
    # window from which we destroy all buttonlist window info.
    set f [frame $state(buttwin)]

    if {![llength $state(-buttonlist)]} {
	# No buttons.
	return
    }

    grid configure $f -row $state(row,buttwin) \
			    -column 0 -columnspan $buttcols -sticky news

    # Set up checkbuttons for controlling substitution:
    set g [frame $f.subst]
    grid configure $g -row 0 -column [expr $buttcols - 1] -sticky news
    set bs [checkbutton $g.bs -text "\\" -fg $CmdButtonColor \
			    -selectcolor $CmdButtonColor \
			    -variable ${NSI}(button,back_subst)]
    set cs [checkbutton $g.cs -text "\[ ]" -fg $CmdButtonColor \
			    -selectcolor $CmdButtonColor \
			    -variable ${NSI}(button,cmd_subst)]
    set vs [checkbutton $g.vs -text "\$" -fg $CmdButtonColor \
			    -selectcolor $CmdButtonColor \
			    -variable ${NSI}(button,var_subst)]
    grid configure $bs -row 0 -column 0 -sticky ew
    grid configure $vs -row 0 -column 1 -sticky ew
    grid configure $cs -row 0 -column 2 -sticky ew

    set row 0
    set col -1
    foreach {label action} $state(-buttonlist) {
	incr col
	if {$col == ($buttcols-1) && $row == 0} {
	    # Skip this spot: we used it for the substitution-buttons.
	    incr row ; set col 0
	} elseif {[string compare $label ""] == 0 &&
					[string compare $action ""] == 0} {
	    # Force new row; don't create button.
	    incr row ; set col 0
	    continue
	} elseif {$col >= $buttcols} {
	    # Need
	    incr row ; set col 0
	}
	set butt $f.b_${row}_${col}
	button $butt -text $label -fg $CmdButtonColor \
		-activeforeground $CmdButtonColor -command \
		[list ${NS}::send_device $instance $action frombutton]
	grid configure $butt -row $row -column $col -sticky nsew
    }
}

proc open_dialog {instance} {
    variable $instance
    upvar 0 $instance state

    set NS [namespace current]
    set NSI [namespace which -variable $instance]

    set toplev .polyterm_open_$instance

    if {[winfo exists $toplev]} {
	wm deiconify $toplev
	center_win $toplev
	return
    }

    toplevel $toplev

    set f [frame $toplev.filespec]
    pack $f -side top -fill x -expand true

    label $f.lab -text "File:"
    pack $f.lab -side left
    entry $f.file -textvariable ${NSI}(path) -width 40
    pack $f.file -side left -fill x -expand true
    button $f.browse -text "Browse" \
		    -command [list ${NS}::poly_getOpenFile $instance]
    pack $f.browse -side left

    set f [frame $toplev.devtype -bd 10]
    pack $f -side top -fill both -expand true

    label $f.lab -text "Type of device:"
    pack $f.lab -side top -anchor w
    set i 0
    foreach {type} [type_info -list ""] {
	radiobutton $f.$i -variable ${NSI}(-devtype) \
					-text $type -value $type
	pack $f.$i -side top -anchor w
	incr i
    }
    radiobutton $f.$i -variable ${NSI}(-devtype) \
				    -text "None of the above" -value ""
    pack $f.$i -side top -anchor w

    set f [frame $toplev.buttons]
    pack $f -side top

    button $f.apply -text "Apply" -command \
		"${NS}::reconfig_and_open_device $instance ; destroy $toplev"
    button $f.cancel -text "Cancel" -command [list destroy $toplev]
    pack $f.apply $f.cancel -side left -fill x

    focus $toplev
    center_win $toplev
    catch {tkwait visibility $toplev}
    catch {grab $toplev}
    catch {tkwait window $toplev}
    grab release $toplev
}

#
# Configure device configuration, then open the device.
#
proc reconfig_and_open_device {instance} {
    variable $instance
    upvar 0 $instance state

    set NS [namespace current]
    set NSI [namespace which -variable $instance]

    # Save the devtype before we re-init...
    set devtype $state(-devtype)
    # Then re-init...
    init_instance $instance 1
    # ...And configure for this devtype again.
    set optlist [list -devtype $devtype]
    if {[catch {config_instance $instance $optlist} result]} {
	tk_dialog .d "Internal Configuration Error" \
		    "Can't configure $state(path) --\
		    there is an error in the device information\
		    for `$devtype': $result" error 0 OK
	return
    }

    open_device $instance
}

#
# Wrapper for tk_getOpenFile -- starts with -initialdir derived from
# the current $state(path) variable.
#
proc poly_getOpenFile {instance} {
    variable $instance
    upvar 0 $instance state

    set optargs {}

    # Find the longest directory substring of the current $state(path),
    # and use that as the starting dir.
    set path $state(path)
    set oldlen -1
    while {[string compare $path ""] && ![file isdirectory $path]} {
	# Entire thing isn't a directory; trim last component and
	# try again.
	set pathlist [lreplace [file split $path] end end]
	if {![llength $pathlist] || [llength $pathlist] == $oldlen} {
	    # No components left, or at least the length didn't change.
	    set path ""
	    break
	}
	set oldlen [llength $pathlist]
	set path [file join $pathlist]
    }
    if {[string compare $path ""]} {
	lappend optargs "-initialdir" $state(path)
    }
    set file [eval tk_getOpenFile $optargs]
    if {[string compare $file ""]} {
	set state(path) $file
    }
    set state(path)
}

proc setcomm {instance} {
    variable $instance
    upvar 0 $instance state	; # Note that state is now an alias for the

    set NS [namespace current]
    set NSI [namespace which -variable $instance]

    set toplev .polyterm_comm_$instance

    if {[winfo exists $toplev]} {
	wm deiconify $toplev
	center_win $toplev
	return
    }

    toplevel $toplev

    # Create some temporary work variables, so that we can cancel easily.
    foreach v {-speed -parity +databits +stopbits} {
	set state(x_$v) $state($v)
    }

    set f [frame $toplev.speed]
    grid configure $f -row 0 -column 0 -sticky n
    set l [label $f.lab -text "Speed"]
    pack $l -side top -fill x
    foreach i {"" 1200 2400 4800 9600 19200 38400 57600 76800 115200 153600} {
	radiobutton $f.$i -variable ${NSI}(x_speed) -text $i -value $i
	pack $f.$i -side top -anchor w
    }

    set f [frame $toplev.parity]
    grid configure $f -row 0 -column 1 -sticky n
    set l [label $f.lab -text "Parity"]
    pack $l -side top -fill x
    foreach {txt val} {"" "" None n Mark m Space s Odd o Even e} {
	radiobutton $f.p_$val -variable ${NSI}(x_parity) -text $txt -value $val
	pack $f.p_$val -side top -anchor w
    }

    set f [frame $toplev.databits]
    grid configure $f -row 0 -column 2 -sticky n
    set l [label $f.lab -text "Data Bits"]
    pack $l -side top -fill x
    foreach bits {"" 5 6 7 8} {
	radiobutton $f.d_$bits -variable ${NSI}(x_databits) \
						-text $bits -value $bits
	pack $f.d_$bits -side top -anchor w
    }

    set f [frame $toplev.stopbits]
    grid configure $f -row 0 -column 3 -sticky n
    set l [label $f.lab -text "Stop Bits"]
    pack $l -side top -fill x
    foreach bits {"" 1 2} {
	radiobutton $f.s_$bits -variable ${NSI}(x_stopbits) \
						-text $bits -value $bits
	pack $f.s_$bits -side top -anchor w
    }

    set f [frame $toplev.buttons]
    grid configure $f -row 1 -column 0 -columnspan 4 -sticky ew
    set cmd { set @NSI@(-speed) $@NSI@(x_speed)
	    set @NSI@(-parity) $@NSI@(x_parity)
	    set @NSI@(+databits) $@NSI@(x_databits)
	    set @NSI@(+stopbits) $@NSI@(x_stopbits)}
    regsub -all @NSI@ $cmd $NSI cmd
    regsub -all @NS@ $cmd $NS cmd
    append cmd "\n${NS}::setmode_ifpossible $instance ; destroy $toplev"
    button $f.apply -text "Apply" -command $cmd
    button $f.cancel -text "Cancel" -command [list destroy $toplev]
    pack $f.apply $f.cancel -side left -fill x

    center_win $toplev
}


proc setstatus {instance openclosed} {
    variable $instance
    upvar 0 $instance state

    set status $state(path)
    if {[string compare $state(name) ""]} {append status " ($state(name))"}
    append status "    $state(modestring)"
    set StatusCmd $state(statuscmd)
    lappend StatusCmd $status
    eval $StatusCmd

    if {[string compare $openclosed "open"] == 0} {
	# Create entry for open device.
	eval $state(statuscolorcmd) black
	$state(filemenu) entryconfigure $state(openitem) -state disabled
	$state(filemenu) entryconfigure $state(closeitem) -state normal
    } else {
	# Create entry for closed device.
	eval $state(statuscolorcmd) grey50
	$state(filemenu) entryconfigure $state(openitem) -state normal
	$state(filemenu) entryconfigure $state(closeitem) -state disabled
    }
}


#
# Invert the value of a boolean element of one of our instance state arrays.
#
proc invert {instance idx} {
    variable $instance
    upvar 0 $instance state
    set state($idx) [expr !$state($idx)]
}

#
# Invoke this if a terminal is destroyed.
#
proc cleanup_terminal {instance} {
    variable Instances
    variable $instance
    upvar 0 $instance state

    catch {close_device $instance}

    catch {unset Instances($instance)}
    catch {unset $instance}
}

proc changefromtext {string newtype} {
    set newstring ""
    switch $newtype {
	binary {
	    foreach c [split $string ""] {
		binary scan $c B8 x
		append newstring $x " "
	    }
	    # Remove the unwanted trailing space added above
	    set newstring [string trimright $newstring]
	}
	octal {
	    foreach c [split $string ""] {
		# binary scan doesn't do octal.  Convert to decimal, then octal.
		binary scan $c c d
		append newstring [format "%o " $d]
	    }
	    # Remove the unwanted trailing space added above
	    set newstring [string trimright $newstring]
	}
	hex {
	    foreach c [split $string ""] {
		binary scan $c H2 x
		append newstring $x  " "
	    }
	    # Remove the unwanted trailing space added above
	    set newstring [string trimright $newstring]
	}
	text {
	    set newstring $string
	}
	default {
	    tk_dialog .d "Unknown Type" \
		    "Don't know how to change text to $newtype" error 0 OK
	    return
	}
    }
    set newstring
}

proc changetotext {string oldtype} {
    set newstring ""
    switch $oldtype {
	binary {
	    foreach word [split $string] {
		append newstring [binary format B8 $word]
	    }
	}
	octal {
	    # binary format doesn't do octal.  Convert to decimal, then text.
	    foreach value [split $string] {
		scan $value %o d
		append newstring [binary format c $d]
	    }
	}
	hex {
	    foreach value [split $string] {
		append newstring [binary format H2 $value]
	    }
	}
	text {
	    set newstring $string
	}
	default {
	    tk_dialog .d "Unknown Type" \
		    "Don't know how to change $oldtype to text" error 0 OK
	    return
	}
    }
    set newstring
}

#
# Either change the selected text's display type, else
# just change future settings.
#
proc changeDisplayedType {instance newtype} {
    variable OurXTags
    variable $instance
    upvar 0 $instance state

    $state(text) configure -state normal
    set Range [$state(text) tag ranges sel]	; # Use selection if it exists.
    if {[llength $Range] == 0} {
	# User didn't select text, so we only record what the future format
	# should be.
	set state(fromdevformat) $newtype
	return
    }

    # Subdivide this range into regions that have distinct OurXTags tags.
    # Proceed in reverse order, so that we can replace text w/o messing up
    # the indexes of the next entries.
    foreach {From To} $Range {break}
    foreach {Row Col} [split $From .] {break}
    foreach xtag $OurXTags {
	set to $To
	while {[llength [set range [$state(text) tag prevrange $xtag $to]]]} {
	    # Found a subrange with this xtag
	    # Note that we didn't bound the 'prevrange' command with
	    # a "from" point, because we want to collect ranges that
	    # start before $Range, but reach into $Range.  So, make an
	    # explicit check here -- if ($to > $From), use the range.
	    foreach {from to} $range {break}
	    foreach {row col} [split $to .] {break}
	    if {$row < $Row || ($row == $Row && $col <= $Col)} {break}

	    changeRangeType $instance $from $to $newtype

	    # Set up for next prevrange.
	    set to $from
	}
    }
    $state(text) configure -state disabled
}

#
# Take the text widget data in the range from..to, and change its display
# type from whatever it currently is to be $newtype (binary, octal, ...).
# Then write the modified data out, replacing its old type tag with
# $newtype.
#
proc changeRangeType {instance from to newtype} {
    variable $instance
    upvar 0 $instance state

    # Find its current display type
    set taglist [$state(text) tag names $from]
    foreach oldtype {binary octal hex text} {
	if {[set type_idx [lsearch -exact $taglist $oldtype]] >= 0} {
	    break
	}
    }

    # Replace its type tag with $newtype, but otherwise preserve the taglist.
    if {$type_idx < 0} {
	# No oldtype; treat it as text.
	set oldtype text
	set newtaglist [concat $taglist $newtype]
    } else {
	set oldtype [lindex $taglist $type_idx]
	set newtaglist [lreplace $taglist $type_idx $type_idx $newtype]
    }

    if {[string compare $oldtype $newtype] == 0} {
	# Already converted
	return
    }

    set oldstring [$state(text) get $from $to]
    if {[string compare $oldtype text]} {
	# Convert oldtype to "text" type, as intermediate form.
	set text [changetotext $oldstring $oldtype]
    } else {
	set text $oldstring
    }
    if {[string compare $newtype text]} {
	# Convert from intermediate text form to final form: 
	set newstring [changefromtext $text $newtype]
    } else {
	set newstring $text
    }
    $state(text) delete $from $to
    $state(text) insert $from $newstring $newtaglist
}


proc confirm_quit {instance} {
    variable Instances
    variable $instance

    if {[tk_dialog .d "Quit?" "Really Quit?" question 0 \
                                            "Don't quit" "Really Quit"]} {
	# Make sure all devices are closed before exiting.
	foreach i [array names Instances] {
	    close_device $i
	}
	exit 0
    }
}

proc subst_append {instance entrybox text} {
    variable $instance
    upvar 0 $instance state
    
    # I think we don't want to do substitution here, because it will
    # be again substituted when the user's typing is sent to the device
    #    set script [list subst]
    #    if {!$state(typed,back_subst)} {
    #	lappend script -nobackslash
    #    }
    #    if {!$state(typed,cmd_subst)} {
    #	lappend script -nocommand
    #    }
    #    if {!$state(typed,var_subst)} {
    #	lappend script -novariable
    #    }
    #    lappend script $text
    #    set text [namespace eval ::$state(-namespace) $script]
    $entrybox insert end $text
}

#
# Return the control character for c
#
proc cntl {c} {
    set c [string toupper $c]
    binary scan $c c n
    binary format c [expr $n - 64]
}

proc open_device {instance} {
    variable $instance
    upvar 0 $instance state

    close_device $instance 1	; # in case it's already open.

    # Create device-specific cmd area (has the text-insertion buttons).
    create_cmdwin $instance

    # Create device-specific buttonlist area (has the command buttons).
    create_buttwin $instance

    setup_delays $instance

    set state(uta) ""
    set state(my_music_addr) -1
    set state(traffic+port)		""
    if {[string compare $state(path) ""] == 0} {
	tk_dialog .d "No device named" \
		    "No device has been specified!" error 0 OK
	return

    } elseif {[regexp "traffic://(.+)" $state(path) junk state(uta)]} {
	# Use music messages.
	if {[string compare $state(-msg) ""] == 0} {
	    tk_dialog .d "Can't use MUSIC Messages" \
		"To use MUSIC, you must specify -msg msgnumber" error 0 OK
	    return
	}
	if {[catch {package require traffic} result]} {
	    tk_dialog .d "Can't use MUSIC Messages" \
		"Failed to find package \"traffic\" -- can't send\
		using MUSIC messages -- $result" error 0 OK
	    return
	}
	if {[catch {traffic::connect -uta $state(uta)} state(handle)]} {
	    tk_dialog .d "Open Failure" \
		"Failed to make MUSIC connection to $state(path):\
			$state(handle)" error 0 OK
	    return
	}

	foreach {canon_name ournm host port a_kind_of_addr} \
			[traffic::canonicalize $state(uta)] {break}
	set state(traffic+port) $host:$port
	foreach {host port myname myaddr} \
			[traffic::traffinfo $state(handle)] {break}
	set state(my_music_addr) $myaddr

	#
	# Set up music message handling.
	#
	traffic::callback [list [namespace current]::handle_music_msg $instance]

	#
	# We exepct messages as broadcasts; request traffic daemon to
	# give us these broadcasts.
	#
	traffic::sendmsg "$state(traffic+port)/0" DESIRED_MSGS \
						int $state(-msg) int -1


    } elseif {[regexp "(.+):(.+)" $state(path) junk host port]} {
	# An open attempt on a socket can block for a very long time.
	# So, we create the socket asynch, and if we can't get
	# the peername w/in a few seconds, we conclude that we can't
	# connect.
	if {[catch {socket -async $host $port} state(handle)]} {
	    tk_dialog .d "Open Failure" \
		"Failed to open $host port $port: $state(handle)" \
		    error 0 OK
	    return
	}
	set w_waitfor [WaitFor $instance "Waiting for socket connection..."]
	set aggregatedelay 0
	while {[catch {fconfigure $state(handle) -peername} result]} {
	    # Couldn't get peername... either we're not connected yet,
	    # or another, more severe error.
	    foreach {type code str} $::errorCode {break}
	    if {[string compare $code ENOTCONN] == 0} {
		# wait 200ms, then try again.
		after 200 [list set ${instance}(open_timeout_scratch) 1 ]
		vwait ${instance}(open_timeout_scratch)
		incr aggregatedelay 200
		if {$aggregatedelay >= $state(timeout)} {
		    destroy $w_waitfor
		    tk_dialog .d "Open Failure" \
		    	"Failed to open $host port $port:\
					couldn't get a socket connection" \
			error 0 OK
		    return
		}
	    } else {
		# Whoops, unexpected error.
		destroy $w_waitfor
		tk_dialog .d "Open Failure" \
		    "Failed to open $host port $port: $code: $str" \
		    error 0 OK
		return
	    }
	}
	destroy $w_waitfor

    } else {
	if {[catch {open $state(path) r+} state(handle)]} {
	    tk_dialog .d "Open Failure" $state(handle) error 0 OK
	    return
	}
    }

    if {$state(my_music_addr) == -1} {

	setup_eol $instance

	setup_serial_mode $instance

	# Not using music messages.
	fileevent $state(handle) readable \
    			[list [namespace current]::read_device $instance]
    }

    setstatus $instance open
}

#
# Convert the user's -discard list (which has been stored in the 
# +delayamount,X elements) into delaychars info.
#
proc setup_delays {instance} {
    variable $instance
    upvar 0 $instance state

    set state(delaychars) ""
    set prefix "+delayamount,"
    set alist [array names state "$prefix*"]
    if {[string length $alist]} {
	set n [string length $prefix]
	foreach arrayname $alist {
	    set delaychar [string index $arrayname $n]
	    append state(delaychars) $delaychar
	}
    }
    # Convert the delaychars string to an order that will work
    # inside a regexp selector list: namely, "]" at the front, and "-"
    # at the end.
    set a ""; set b ""
    if {[regexp "(.*)\](.*)" $state(delaychars) junk a b]} {
	set state(delaychars) "\]$a$b"
    }
    set a ""; set b ""
    if {[regexp "(.*)-(.*)" $state(delaychars) junk a b]} {
	set state(delaychars) "$a$b-"
    }
}

#
# Configure the end-of-line handling and channel encoding.
#
proc setup_eol {instance} {
    variable $instance
    upvar 0 $instance state

    #
    # Use the same end-of-line translation for sending and receiving.
    #
    set translation $state(-termination)

    fconfigure $state(handle) -blocking 0 -translation $translation
    # Treat the device as raw bytes.
    if {[info tclversion] >= 8.1} {
	fconfigure $state(handle) -encoding binary
    }
}

#
# Gets serial settings for a handle; returns as list {baud, parity, data, stop}
# Returns "" if the handle doesn't accept -mode option.
# Generates an error if the handle is n.g.
#
proc get_serial {handle} {
    if {[catch {fconfigure $handle} result]} {
	# Not a valid handle.
	return -code error $result
    }
    if {[catch {fconfigure $handle -mode} result]} {
	# Not a serial device.
	return {}
    }
    split $result ,
}

proc setmode_ifpossible {instance} {
    variable $instance
    upvar 0 $instance state
    
    if {[string compare $state(handle) ""] == 0} {
	# Device not open; can't set mode
	return
    }
    setup_serial_mode $instance
    setstatus $instance open
}

proc setup_serial_mode {instance} {
    variable $instance
    upvar 0 $instance state

    # If this is a serial device, set mode accordingly.
    if {[catch {get_serial $state(handle)} mode]} {
	# Not valid handle
	return -code error $mode
    }
    if {![llength $mode]} {
	# Not a serial device.  Error if user specified any serial options.
	set state(modestring) ""
	if {[string compare $state(+databits) ""] ||
		[string compare $state(+stopbits) ""] ||
		    [string compare $state(-parity) ""] ||
			[string compare $state(-speed) ""]} {
	    tk_dialog .d "Configuration Error" "This device can't use the\
	    	-bits, -parity, or -speed options.  Ignoring your\
		settings." error 0 OK
	    return
	}
    } else {
	#
	# Serial device.  Use any user-specified values.
	#
	foreach {baud parity data stop} $mode {break}
	if {[string compare $state(-speed) ""]} { set baud $state(-speed) }
	if {[string compare $state(-parity) ""]} { set parity $state(-parity) }
	if {[string compare $state(+databits) ""]} { set data $state(+databits)}
	if {[string compare $state(+stopbits) ""]} { set stop $state(+stopbits)}
	set mode "$baud,$parity,$data,$stop"
	if {[catch {fconfigure $state(handle) -mode $mode} result]} {
	    tk_dialog .d "Serial Line Setting Error" \
	    	"Failed to set baud/parity/databits/stopbits\
			to $mode: $result" error 0 OK

	    # Ensure modestring is accurate, regardless of what the
	    # fconfigure did...
	    foreach {baud parity data stop} [get_serial $state(handle)] {break}
	    set state(modestring) "$baud $data[string toupper $parity]$stop"
	    return
	}
	set state(modestring) "$baud $data[string toupper $parity]$stop"
    }
}

proc close_device {instance {destroy 0}} {
    variable $instance
    upvar 0 $instance state

    if {$destroy} {
	# Close device-specific cmd area (has the text-insertion buttons).
	if {[winfo exists $state(cmdwin)]} { destroy $state(cmdwin) }

	# Close device-specific buttonlist area (has the command buttons).
	if {[winfo exists $state(buttwin)]} { destroy $state(buttwin) }
    }

    # Close the device
    if {$state(my_music_addr) == -1} {
	if {[string compare $state(handle) ""]} {
	    catch {
		close $state(handle);
		fileevent $state(handle) readable {}
	    }
	}
	set state(handle) ""
    } else {
	traffic::disconnect $state(uta)
    }
    setstatus $instance closed
}

#
# Evaluate from a tcl-code entry box, and then reset its status
#
proc eval_code_and_reset {instance widget} {
    variable $instance
    upvar 0 $instance state

    eval_code $instance $state(codeentry)
    $widget delete 0 end
}

#
# Evaluate user code in the user's namespace.
#
proc eval_code {instance code} {
    variable $instance
    upvar 0 $instance state

    if {!$state(-catchputs)} {
	# No special puts games.
	set rename_puts 0
    } else {
	# Catch puts to stdout and redirect to text widget
	set rename_puts 1
	set body [info body pt_puts]
	regsub -all @NS@ $body [namespace current] body
	set script {
	    rename ::puts ::true_puts
	    proc puts {args} {set instance $instance ; $body}
	}
	set script [subst $script]
	set prescript [list set [namespace current]::this $instance]
	namespace eval ::$state(-namespace) \
		[concat $prescript ";" $script]
    }

    # Now execute the user's code.
    catch {namespace eval ::$state(-namespace) $code} result

    # And restore puts, if it was renamed.
    if {$rename_puts} {
	catch {rename ::puts ""}	; # in case user's namespace was ::
	rename ::true_puts ::puts
    }

    # And stuff the output, if there was any.
    if {[string compare $result ""]} {
	stuff_text $instance "$result\n" tclcode
    }
}

#
# Send data from a command entry box, and then reset its status.
#
proc send_and_reset {instance widget {tag ""}} {
    variable $instance
    upvar 0 $instance state

    if {[send_device $instance $state(cmdentry) $tag]} {
	$widget delete 0 end
	set state(inquote) 0
	set state(cmdhistidx) 0
    }
}

proc senddev {text {tag tclcode}} {
    set this [set [namespace current]::this]
    send_device $this $text $tag
}

proc send_device {instance text {tag ""}} {
    variable $instance
    upvar 0 $instance state
    
    set NS [namespace current]

    if {[string compare $state(handle) ""] == 0} {
	tk_dialog .d "No Device Open" \
		    "Can't write; device is not open" error 0 OK
	return 0
    }
    switch -exact -- $tag {
	frombutton {
		    set script [list subst]
		    if {!$state(button,back_subst)} {
			lappend script -nobackslash
		    }
		    if {!$state(button,cmd_subst)} {
			lappend script -nocommand
		    }
		    if {!$state(button,var_subst)} {
			lappend script -novariable
		    }
		    lappend script $text
		    set text [namespace eval ::$state(-namespace) $script]
	}
	usertyped {
		    if {[string compare $text ""]} {
			lappend state(cmd,history) $text
		    }
		    set script [list subst]
		    if {!$state(typed,back_subst)} {
			lappend script -nobackslash
		    }
		    if {!$state(typed,cmd_subst)} {
			lappend script -nocommand
		    }
		    if {!$state(typed,var_subst)} {
			lappend script -novariable
		    }
		    lappend script $text
		    set text [namespace eval ::$state(-namespace) $script]
	}
    }
    # Now split the text at the delay characters, and insert a suitable
    # delay at each one.
    if {[string length $state(delaychars)] == 0} {
	# no delay chars
	set textlist [list $text]
	set splitlist ""
    } else {
	set textlist [split $text "$state(delaychars)"]
	regsub -all "\[^$state(delaychars)]" $text "" str
	set splitlist [split $str ""]
    }
    foreach line $textlist char $splitlist {
	if {$state(my_music_addr) != -1} {
	    # use MUSIC messages
	    traffic::sendmsg $state(uta) $state(-msg) buf1 $state(-devnum) \
	    							str $line
	} else {
	    if {$state(+use_termination)} {
		puts $state(handle) $line
	    } else {
		puts -nonewline $state(handle) $line
	    }
	    flush $state(handle)
	}
	# Stuff the text into the display, too
	stuff_text $instance "$line\n" $tag

	# Depending on where the delays are set, there can be a trailing
	# text element w/o a following delay, so char is empty.  Thus only
	# delay if there really is a character and its delay amount is > 0.
	if {[string length $char] && $state(+delayamount,$char) > 0} {
	    # do this funny after+vwait so that we don't stop all processing.
	    after $state(+delayamount,$char) [list \
	    	set ${NS}::${instance}(delayhelper) 1 ]
	    vwait ${NS}::${instance}(delayhelper)
	}
    }
    return 1
}

proc stuff_text {instance text {tag ""} {mustbenewline 0}} {
    variable $instance
    upvar 0 $instance state

    $state(text) configure -state normal
    if {$mustbenewline} {
	# Ensure that this text begins on a new line.
	# The character before "end" is <line>.<charposition>, ie
	# <line>.0 if at beginning of line.
	if {![regexp {\.0$} [$state(text) index "end-1c"]]} {
	    # Not at start of line.
	    $state(text) insert end "\n"
	}
    }
    if {$state(-crnl)} {
	# Elide \r\n; replace with just \n.
	regsub -all "\r\n" $text "\n" text
	regsub -all "\r" $text "\n" text
    }
    if {[string compare $tag ""]} {
	$state(text) insert end $text $tag
    } else {
	$state(text) insert end $text
    }
    $state(text) yview end
    $state(text) configure -state disabled
}

proc read_device {instance} {
    variable $instance
    upvar 0 $instance state

    if {[eof $state(handle)]} {
	close_device $instance
	tk_dialog .d "Device is Not Open" \
		    "The device is not currently open" error 0 OK
	set state(handle) ""
	return
    }
    set string [read $state(handle)]
    if {[string compare $string ""]} {
	append state(memory) $string
	if {[string compare $state(fromdevformat) text]} {
	    set string [changefromtext $string $state(fromdevformat)]
	}
	stuff_text $instance $string [list fromdevice $state(fromdevformat)]
    }
}

proc handle_music_msg {instance N} {
    variable $instance
    upvar 0 $instance state

    if {[eof $state(handle)]} {
	close_device $instance
	tk_dialog .d "Device is Not Open" \
		    "The device is not currently open" error 0 OK
	set state(handle) ""
	traffic::discard $N
	return
    }

    set msg [traffic::msg $N msg]
    set from [traffic::msg $N from]
    set seq_num [traffic::msg $N seq_num]
    set data1 [traffic::msg $N data1]
    set buf1 [lindex [traffic::msg $N buf] 0]

    if {$msg != $state(-msg)} {
	if {$from == 0 && $msg == $traffic::Definition(R_DESIRED_MSGS)} {
	    ; # expected response to DESIRED_MSGS; ignore.
	} elseif {$from == 0 && $msg == $traffic::Definition(NO_DELIVERY)} {
	    stuff_text $instance \
		"can't send message to traffic client $state(uta)" fromunknown
	} else {
	    stuff_text $instance \
	    	"Unknown message $msg from $from; seq #$seq_num" fromunknown 
	}
	traffic::discard $N
	return
    } elseif {$buf1 != $state(-devnum)} {
	# Ignore messages that aren't regarding the device of interest
	traffic::discard $N
    } elseif {$data1 == $state(my_music_addr)} {
	# Ignore messages that are simply rebroadcasts of what we sent.
	traffic::discard $N
	return
    } elseif {$data1 == -1} {
	set type fromdevice
    } else {
	set type fromunknown
    }

    set text [traffic::body $N "A*"]
    traffic::discard $N
    if {[string compare $text ""]} {
	append state(memory) $text
	if {[string compare $type "fromdevice"] == 0} {
	    if {[string compare $state(fromdevformat) text]} {
		set string [changefromtext $text $state(fromdevformat)]
	    }
	    stuff_text $instance $text [list $type $state(fromdevformat)]
	} else {
	    stuff_text $instance $text $type
	}
    }
}

#
# Centering a window on the screen.
#
proc center_win {w} {
    # The following is taken directly from the tk_dialog widget.

    # Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw $w
    update idletasks
    set x [expr {[winfo screenwidth $w]/2 -
		[winfo reqwidth $w]/2  - [winfo vrootx [winfo parent $w]]}]
    set y [expr {[winfo screenheight $w]/2 -
		[winfo reqheight $w]/2  - [winfo vrooty [winfo parent $w]]}]
    wm geom $w +$x+$y
    wm deiconify $w
}

#
# Pop up a toplevel which does a focus grab and displays a message.
# Update the message dots every second.   Return the toplevel name;
# when the toplevel is destroyed, stop the message updates.
#
#
proc WaitFor {instance {message "Wait"} {grab 1} {maxdots 10}} {
    variable $instance
    upvar 0 $instance state

    set NS [namespace current]
    set NSI ${NS}::$instance

    set toplev .polyterm_waitfor_$instance
    set varname ${NSI}(WaitFor)

    if {![winfo exists $toplev]} {
	# It's a new waitfor command.
	toplevel $toplev
	wm title $toplev "Waiting..."
	label $toplev.l1 -text $message
	label $toplev.l2 -textvariable $varname
	pack $toplev.l1 -side top
	pack $toplev.l2 -side top -anchor w
	# Cancel the [after] update when the widget is destroyed.
	set script {
	    if {[string compare %W $toplev] == 0} {
		catch {after cancel \$${NSI}($toplev,WaitFor)}
	    }
	    catch {grab release $toplev}
	}
	bind $toplev <Destroy> [subst -nocommand $script]
	focus $toplev
	center_win $toplev
	catch {tkwait visibility $toplev}
	catch {grab $toplev}
    } else {
	# Still waiting...
	$toplev.l1 configure -text $message
	if {[string length [set $varname]] >= $maxdots} {
	    set $varname .
	} else {
	    append $varname .
	}
    }
    set state($toplev,WaitFor) [after 1000 \
    			[list ${NS}::WaitFor $instance $message]]
    return $toplev
}

#
# If $action is -list, return list of all known types; ignore the $type.
# If $action is -canonical, return canonical name for this device type;
#	generates an error if $type doesn't exist.
# If $action is -get, return config list for this device type;
#	generates an error if $type doesn't exist.
# Otherwise error.
#
proc type_info {action type} {

    set TypeInfo(Galil) {
	-termination crlf
	-upcase yes
	-buttonlist {}
	-insertlist {}
    }

    set TypeInfo(HP34970A) {
	-termination lf
	-discard {\n 100}
	-buttonlist {
	    "Communications\nCheck:\nAsk Version"       "SYST:VERS?"
	    "Inquire:\n# Points\nIn Memory"             "DATA:POINTS?"
	    "Inquire:\nScanlist"                        "ROUTE:SCAN?"
	    "Inquire:\nTrigger Info"
	    	"TRIGGER:SOURCE?\nTRIGGER:TIM?\nTRIGGER:COUNT?"
	}
	-insertlist {}
    }

    set name "Lamp Strip"
    set TypeInfo($name) {
	-termination none
	-discard {\t 100}
	-buttonlist {
	        "Inquire\nState" "A"
		"Flash\nLights" "C"
	}
	-insertlist {}
    }

    if {[string compare $action -list] == 0} {
	# just return the known TypeInfo's
	return [array names TypeInfo]

    }

    # If we get here, we have to find info on a type.
    set match 0

    #
    # First try exact match...
    #
    if {[info exists TypeInfo($type)]} {
	set match 1
    }

    if {!$match} {
	#
	# Next try case-insensitive glob match, treating user's type
	# as leading substring of desired type.
	#
	set TypeList [array names TypeInfo]
	set typelist {}
	foreach t $TypeList {
	    lappend typelist [string tolower $t]
	}
	if {[set i [lsearch -glob $typelist $type*]] >= 0} {
	    set match 1
	    set type [lindex $TypeList $i]
	}
    }

    if {!$match} {
	#
	# Next try ignoring whitespace, and repeating the glob match.
	#
	set typelistx {}
	foreach t $typelist {
	    regsub -all "\[ \t\r\n]" $t "" t
	    lappend typelistx $t
	}
	if {[set i [lsearch -glob $typelistx $type*]] >= 0} {
	    set match 1
	    set type [lindex $TypeList $i]
	}
    }

    if {!$match} {
	#
	# Sorry, no such type.
	#
	return -code error "No match to device type `$type'"
    }

    if {[string compare $action -canonical] == 0} {
	return $type
    } elseif {[string compare $action -get] == 0} {
	return $TypeInfo($type)
    } else {
	return -code error "Don't know how to handle action $action"
    }
}

} ; # End polyterm namespace
